# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import retrying

from oslo_log import log as logging

from ironic_oneviewd.conf import CONF
from ironic_oneviewd.node_manager.manage import NodeManager

LOG = logging.getLogger(__name__)


def do_manage_ironic_nodes():
    """Show a list of OneView servers to be created as nodes in Ironic."""
    node_manager = NodeManager()
    retry_interval_in_ms = CONF.DEFAULT.retry_interval * 1000

    @retrying.retry(wait_fixed=retry_interval_in_ms)
    def execute():
        try:
            node_manager.pull_ironic_nodes()
        except Exception as ex:
            LOG.error(ex)
        raise Exception("Continue trying...")
    execute()



# Copyright 2016 Hewlett Packard Enterprise Development LP.
# Copyright 2016 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

CONF = cfg.CONF

opts = [
    cfg.StrOpt('auth_url',
               help='Authentication URL for OpenStack services.'),
    cfg.StrOpt('username',
               help='Name of the OpenStack user.'),
    cfg.StrOpt('password',
               secret=True,
               help='OpenStack user password.'),
    cfg.StrOpt('region_name',
               help='Name of the OpenStack Keystone region.'),
    cfg.BoolOpt('insecure',
                default=False,
                help='Allow insecure SSL (no cert verification).'),
    cfg.StrOpt('cacert',
               help='(Optional) Path to CA certificate bundle file.'),
    cfg.StrOpt('cert',
               help='(Optional) Path to cert file.'),
    cfg.StrOpt('deploy_kernel_id',
               help='Ironic deploy kernel image uuid.'),
    cfg.StrOpt('deploy_ramdisk_id',
               help='Ironic deploy remdisk image uuid.'),
    cfg.StrOpt('endpoint_type',
               default='publicURL',
               help='(Optional) OpenStack endpoint type.'),
    cfg.StrOpt('ironic_driver',
               choices=['iscsi_pxe_oneview, agent_pxe_oneview, fake_oneview'],
               help='Ironic OneView driver for node create. '),
    cfg.StrOpt('tenant_id',
               help='ID of the OpenStack tenant. '
                    '(deprecated in favour of project_id)'),
    cfg.StrOpt('tenant_name',
               help='Name of the OpenStack tenant. '
                    '(deprecated in favour of project_name)'),
    cfg.StrOpt('project_id',
               help='ID of the OpenStack project.'),
    cfg.StrOpt('project_name',
               help='Name of the OpenStack project.'),
    cfg.StrOpt('user_domain_id',
               help='ID of a domain the user belongs to.'),
    cfg.StrOpt('user_domain_name',
               help='Name of a domain the user belongs to.'),
    cfg.StrOpt('project_domain_id',
               help='ID of a domain the project belongs to.'),
    cfg.StrOpt('project_domain_name',
               help='ID of a domain the project belongs to.'),
    cfg.BoolOpt('inspection_enabled',
                default=False,
                help='Run inspection on nodes when any hardware property is '
                     'missing.')
]


def register_opts(conf):
    conf.register_opts(opts, group='openstack')



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from ironicclient import client as ironic_client

from oslo_log import log as logging
from oslo_utils import importutils

from ironic_oneviewd.conf import CONF
from ironic_oneviewd import exceptions
from ironic_oneviewd.openstack.common._i18n import _

client = importutils.try_import('oneview_client.client')
oneview_states = importutils.try_import('oneview_client.states')
oneview_exceptions = importutils.try_import('oneview_client.exceptions')

REQUIRED_ON_PROPERTIES = {
    'server_hardware_type_uri': _("Server Hardware Type URI Required."),
    'server_profile_template_uri': _("Server Profile Template URI required"),
}

REQUIRED_ON_EXTRAS = {
    'server_hardware_uri': _("Server Hardware URI. Required."),
}

IRONIC_API_VERSION = 1

LOG = logging.getLogger(__name__)


def get_ironic_client():
    """Generates an instance of the Ironic client.

    This method creates an instance of the Ironic client using the OpenStack
    credentials from config file and the imported ironicclient library.

    :returns: an instance of the Ironic client
    """
    daemon_kwargs = {
        'os_username': CONF.openstack.username,
        'os_password': CONF.openstack.password,
        'os_auth_url': CONF.openstack.auth_url,
        'os_region_name': CONF.openstack.region_name,
        'insecure': CONF.openstack.insecure,
        'os_cacert': CONF.openstack.cacert,
        'os_cert': CONF.openstack.cert,
        'os_endpoint_type': CONF.openstack.endpoint_type,
        'os_project_id': CONF.openstack.project_id,
        'os_project_name': CONF.openstack.project_name,
        'os_tenant_id': CONF.openstack.tenant_id,
        'os_tenant_name': CONF.openstack.tenant_name,
        'os_user_domain_id': CONF.openstack.user_domain_id,
        'os_user_domain_name': CONF.openstack.user_domain_name,
        'os_project_domain_id': CONF.openstack.project_domain_id,
        'os_project_domain_name': CONF.openstack.project_domain_name,
        'os_ironic_api_version': '1.22'
    }

    LOG.debug("Using OpenStack credentials specified in the configuration "
              "file to get Ironic Client")

    return ironic_client.get_client(IRONIC_API_VERSION, **daemon_kwargs)


def get_oneview_client():
    """Generates an instance of the OneView client.

    Generates an instance of the OneView client using the imported
    oneview_client library.

    :returns: an instance of the OneView client
    """
    oneview_client = client.Client(
        manager_url=CONF.oneview.manager_url,
        username=CONF.oneview.username,
        password=CONF.oneview.password,
        allow_insecure_connections=CONF.oneview.allow_insecure_connections,
        tls_cacert_file=CONF.oneview.tls_cacert_file,
        max_polling_attempts=CONF.oneview.max_polling_attempts,
        audit_enabled=CONF.oneview.audit_enabled,
        audit_map_file=CONF.oneview.audit_map_file,
        audit_output_file=CONF.oneview.audit_output_file
    )
    return oneview_client


def verify_node_properties(node):
    properties = node.properties.get('capabilities', '')
    for key in REQUIRED_ON_PROPERTIES:
        if key not in properties:
            raise exceptions.MissingParameterValue(
                _("Missing the following OneView data in node's "
                  "properties/capabilities: %s.") % key
            )

    return properties


def verify_node_extra(node):
    extra = node.extra or {}
    for key in REQUIRED_ON_EXTRAS:
        if not extra.get(key):
            raise exceptions.MissingParameterValue(
                _("Missing the following OneView data in node's extra: %s.")
                % key
            )

    return extra


def capabilities_to_dict(capabilities):
    """Parse the capabilities string into a dictionary

    :param capabilities: the node capabilities as a formatted string
    :raises: InvalidParameterValue if capabilities is not an string or has
             a malformed value
    """
    capabilities_dict = {}
    if capabilities:
        if not isinstance(capabilities, six.string_types):
            raise exceptions.InvalidParameterValue(
                _("Value of 'capabilities' must be string. Got %s")
                % type(capabilities))
        try:
            for capability in capabilities.split(','):
                key, value = capability.split(':')
                capabilities_dict[key] = value
        except ValueError:
            raise exceptions.InvalidParameterValue(
                _("Malformed capabilities value: %s") % capability
            )
    return capabilities_dict


def dynamic_allocation_enabled(node):
    flag = node.driver_info.get('dynamic_allocation')
    if flag:
        if str(flag).lower() == 'true':
            return True
        elif str(flag).lower() == 'false':
            return False
        else:
            msg = (("Invalid dynamic_allocation parameter value "
                    "'%(flag)s' in node's %(node_uuid)s driver_info. "
                    "Valid values are booleans true or false.") %
                   {"flag": flag, "node_uuid": node.uuid})
            raise exceptions.InvalidParameterValue(msg)
    return False


def get_node_info_from_node(node):
    capabilities_dict = capabilities_to_dict(
        node.properties.get('capabilities', '')
    )
    driver_info = node.driver_info
    oneview_info = {
        'server_hardware_uri':
            driver_info.get('server_hardware_uri'),
        'server_hardware_type_uri':
            capabilities_dict.get('server_hardware_type_uri'),
        'enclosure_group_uri':
            capabilities_dict.get('enclosure_group_uri'),
        'server_profile_template_uri':
            capabilities_dict.get('server_profile_template_uri') or
            driver_info.get('server_profile_template_uri')
    }
    return oneview_info


def node_has_hardware_propeties(node):
    properties = [node.properties.get("memory_mb"),
                  node.properties.get("cpu_arch"),
                  node.properties.get("local_gb"),
                  node.properties.get("cpus")]
    return all(properties)


def server_profile_template_uri_from_node(node):
    node_capabilities = capabilities_to_dict(
        node.properties.get('capabilities')
    )
    node_server_profile_template_uri = node_capabilities.get(
        'server_profile_template_uri'
    )
    return node_server_profile_template_uri


def server_hardware_uri_from_node(node):
    return node.driver_info.get('server_hardware_uri')


def server_hardware_uuid_from_node(node):
    uri = server_hardware_uri_from_node(node)
    return uuid_from_uri(uri)


def uuid_from_uri(uri):
    return uri.split("/")[-1]



# Copyright 2016 Hewlett Packard Enterprise Development LP.
# Copyright 2016 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ironic_oneviewd.conf

_opts = [
    ('DEFAULT', ironic_oneviewd.conf.default.opts),
    ('openstack', ironic_oneviewd.conf.openstack.opts),
    ('oneview', ironic_oneviewd.conf.oneview.opts)
]


def list_opts():
    """Return a list of oslo.config options available in Ironic-OneView Daemon.

    The returned list includes all oslo.config options. Each element of
    the list is a tuple. The first element is the name of the group, the
    second element is the options.

    The function is discoverable via the 'ironic-oneviewd' entry point under
    the 'oslo.config.opts' namespace.

    The function is used by Oslo sample config file generator to discover the
    options.

    :returns: a list of (group, options) tuples
    """
    return _opts



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from concurrent import futures
from oslo_log import log as logging

from ironic_oneviewd.conf import CONF
from ironic_oneviewd import exceptions
from ironic_oneviewd import facade
from ironic_oneviewd import utils

LOG = logging.getLogger(__name__)

ENROLL_PROVISION_STATE = 'enroll'
MANAGEABLE_PROVISION_STATE = 'manageable'
AVAILABLE_PROVISION_STATE = 'available'
INSPECTION_FAILED_PROVISION_STATE = 'inspect failed'
ONEVIEW_PROFILE_APPLIED = 'ProfileApplied'
IN_USE_BY_ONEVIEW = 'is already in use by OneView.'

SUPPORTED_DRIVERS = [
    'agent_pxe_oneview',
    'iscsi_pxe_oneview',
    'fake_oneview'
]

ACTION_STATES = [ENROLL_PROVISION_STATE,
                 MANAGEABLE_PROVISION_STATE,
                 INSPECTION_FAILED_PROVISION_STATE]


class NodeManager(object):
    def __init__(self):
        self.facade = facade.Facade()
        self.max_workers = CONF.DEFAULT.rpc_thread_pool_size
        self.executor = futures.ThreadPoolExecutor(
            max_workers=self.max_workers
        )

    def pull_ironic_nodes(self):
        ironic_nodes = self.facade.get_ironic_node_list()

        nodes = [node for node in ironic_nodes
                 if node.driver in SUPPORTED_DRIVERS
                 if node.provision_state in ACTION_STATES
                 if node.maintenance is False]

        if nodes:
            LOG.info(
                "%(nodes)s Ironic nodes has been taken." %
                {"nodes": len(nodes)}
            )
            self.executor.map(self.manage_node_provision_state, nodes)

    def manage_node_provision_state(self, node):
        if node.provision_state == ENROLL_PROVISION_STATE:
            self.take_enroll_state_actions(node)
        elif node.provision_state == MANAGEABLE_PROVISION_STATE:
            self.take_manageable_state_actions(node)
        elif node.provision_state == INSPECTION_FAILED_PROVISION_STATE:
            self.take_inspect_failed_state_actions(node)

    def take_enroll_state_actions(self, node):
        LOG.info(
            "Taking enroll state actions for node %(node)s." %
            {'node': node.uuid}
        )

        if utils.dynamic_allocation_enabled(node):
            try:
                self.apply_node_port_conf_for_dynamic_allocation(node)
            except exceptions.NodeAlreadyHasPortForThisMacAddress as ex:
                LOG.warning(six.text_type(ex))
            except exceptions.NoBootableConnectionFoundException as ex:
                LOG.warning(six.text_type(ex))
                return
        else:
            try:
                self.apply_server_profile(node)
            except (
                    exceptions.NodeAlreadyHasServerProfileAssignedException
            ) as ex:
                LOG.warning(six.text_type(ex))
            except exceptions.ServerProfileApplicationException as ex:
                LOG.warning(six.text_type(ex))
                return

            try:
                self.apply_node_port_configuration(node)
            except exceptions.NodeAlreadyHasPortForThisMacAddress as ex:
                LOG.warning(six.text_type(ex))
            except exceptions.NoBootableConnectionFoundException as ex:
                LOG.warning(six.text_type(ex))
                return

        try:
            self.facade.set_node_provision_state(node, 'manage')
        except Exception as e:
            LOG.error(e.message)

    def take_manageable_state_actions(self, node):
        LOG.info(
            "Taking manageable state actions for node %(node)s." %
            {'node': node.uuid}
        )

        if utils.dynamic_allocation_enabled(node):
            if (CONF.openstack.inspection_enabled and
                    not utils.node_has_hardware_propeties(node)):
                self.facade.set_node_provision_state(node, 'inspect')
                return
            elif (not CONF.openstack.inspection_enabled and
                    not utils.node_has_hardware_propeties(node)):
                LOG.warning(
                    "Node %(node)s has missing hardware properties and "
                    "Inspection is not enabled." % {'node': node.uuid}
                )
        try:
            self.facade.set_node_provision_state(node, 'provide')
        except Exception as e:
            LOG.error(e.message)

    def take_inspect_failed_state_actions(self, node):
        if (utils.dynamic_allocation_enabled(node) and
                node.last_error and (IN_USE_BY_ONEVIEW in node.last_error)):
            LOG.info(
                "Inspection failed on node %(node)s due to machine being in "
                "use by OneView. Moving it back to manageable state." %
                {'node': node.uuid}
            )

            self.facade.set_node_provision_state(node, 'manage')

        else:
            LOG.warning(
                "Inspection failed on node %(node)s." % {'node': node.uuid}
            )

    def server_hardware_has_server_profile_fully_applied(self, node):
        server_hardware_uuid = utils.server_hardware_uuid_from_node(node)
        server_hardware_state = self.facade.get_server_hardware_state(
            server_hardware_uuid
        )
        return server_hardware_state == ONEVIEW_PROFILE_APPLIED

    def apply_server_profile(self, node):
        server_profile_uri = self.server_profile_uri_from_node(
            node
        )

        node_info = utils.get_node_info_from_node(node)
        server_hardware = self.facade.get_server_hardware(node_info)
        server_profile_template_uri = node_info.get(
            'server_profile_template_uri'
        )
        server_profile_template_uuid = utils.uuid_from_uri(
            server_profile_template_uri
        )

        server_profile_name = "Ironic [%s]" % (node.uuid)

        if server_profile_uri is None:
            try:
                server_profile_uri = (
                    self.facade.generate_and_assign_sp_from_spt(
                        server_profile_name,
                        server_hardware.uuid,
                        server_profile_template_uuid
                    )
                )
            except Exception:
                raise exceptions.ServerProfileApplicationException(node)
        else:
            raise exceptions.NodeAlreadyHasServerProfileAssignedException(
                node,
                server_profile_uri
            )

    def apply_node_port_configuration(self, node):
        if not self.server_hardware_has_server_profile_fully_applied(node):
            LOG.error(
                "The Server Profile application is not "
                "finished yet for node %(node)s." %
                {'node': node.uuid}
            )
            return

        node_info = utils.get_node_info_from_node(node)

        assigned_server_profile_uri = self.server_profile_uri_from_node(
            node
        )
        server_profile = self.facade.get_server_profile_assigned_to_sh(
            node_info
        )
        primary_boot_connection = None

        if server_profile.connections:
            for connection in server_profile.connections:
                boot = connection.get('boot')
                if (boot is not None and
                   boot.get('priority').lower() == 'primary'):
                    primary_boot_connection = connection

            if primary_boot_connection is None:
                raise exceptions.NoBootableConnectionFoundException(
                    assigned_server_profile_uri
                )
            mac = primary_boot_connection.get('mac')
        else:
            server_hardware_uuid = utils.server_hardware_uuid_from_node(node)
            mac = self.facade.get_server_hardware_mac(server_hardware_uuid)

        return self.get_a_port_to_apply_to_node(node, mac)

    def apply_node_port_conf_for_dynamic_allocation(self, node):
        server_hardware_uuid = utils.server_hardware_uuid_from_node(node)
        mac = self.facade.get_server_hardware_mac(server_hardware_uuid)
        return self.get_a_port_to_apply_to_node(node, mac)

    def server_profile_uri_from_node(self, node):
        node_info = utils.get_node_info_from_node(node)
        server_profile_uri = None
        try:
            server_profile = self.facade.get_server_profile_assigned_to_sh(
                node_info
            )

            if node.uuid in server_profile.name:
                return server_profile.uri
            else:
                return server_profile_uri

        except Exception:
            return server_profile_uri

    def is_bootable(self, node, mac):
        node_info = utils.get_node_info_from_node(node)
        server_hardware = self.facade.get_server_hardware(node_info)

        # DL hardware don't have managed networking.
        if not server_hardware.enclosure_group_uri:
            return True

        for server_hardware in server_hardware.port_map['deviceSlots']:
            for physicalPorts in server_hardware['physicalPorts']:
                if mac and mac.upper() == physicalPorts['mac'].upper():
                    return True
        return False

    def set_local_link_connection(self, node, mac):
        local_link_connection = {}
        if node.driver_info.get('use_oneview_ml2_driver'):
            server_hardware_id = utils.server_hardware_uuid_from_node(node)
            switch_info = (
                '{"server_hardware_id": "%(server_hardware_id)s", '
                '"bootable": "%(bootable)s"}') % {
                    'server_hardware_id': server_hardware_id,
                    'bootable': self.is_bootable(node, mac)}
            local_link_connection = {
                "switch_id": "01:23:45:67:89:ab",
                "port_id": "",
                "switch_info": switch_info
            }
        return local_link_connection

    def get_a_port_to_apply_to_node(self, node, mac):
        port_list_by_mac = self.facade.get_port_list_by_mac(mac)
        if not port_list_by_mac:
            return self.apply_port_to_node(node, mac)
        else:
            port_obj = self.facade.get_port(port_list_by_mac[0].uuid)
            if port_obj.node_uuid != node.uuid:
                return self.apply_port_to_node(node, mac)
            else:
                raise exceptions.NodeAlreadyHasPortForThisMacAddress(mac)

    def apply_port_to_node(self, node, mac):
        if node.driver_info.get('use_oneview_ml2_driver'):
            local_link_connection = self.set_local_link_connection(node, mac)
            return self.facade.create_node_port(node.uuid, mac,
                                                local_link_connection)
        self.facade.create_node_port(node.uuid, mac)



# Copyright 2016 Hewlett Packard Enterprise Development LP.
# Copyright 2016 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

CONF = cfg.CONF

opts = [
    cfg.StrOpt('manager_url',
               help='URL where OneView is available.'),
    cfg.StrOpt('username',
               help='OneView username to be used.'),
    cfg.StrOpt('password',
               secret=True,
               help='OneView password to be used.'),
    cfg.BoolOpt('allow_insecure_connections',
                default=False,
                help='Option to allow insecure connection with OneView.'),
    cfg.StrOpt('tls_cacert_file',
               help='(Optional) Path to CA certificate.'),
    cfg.IntOpt('max_polling_attempts',
               default=12,
               help='Max connection retries to check changes on OneView.'),
    cfg.BoolOpt('enable_periodic_tasks',
                default=True,
                help='(Optional) Whether to enable the periodic tasks for '
                     'OneView driver be aware when OneView hardware resources '
                     'are taken and released by Ironic or OneView users '
                     'and proactively manage nodes in clean fail state '
                     'according to Dynamic Allocation model of hardware '
                     'resources allocation in OneView.'),
    cfg.IntOpt('periodic_check_interval',
               default=300,
               help='Period (in seconds) for periodic tasks to be '
                    'executed when enable_periodic_tasks is True.'),
    cfg.BoolOpt('audit_enabled',
                default=False,
                help='(Optional) Enable auditing of OneView API requests.'),
    cfg.StrOpt('audit_map_file',
               help='Path to map file for OneView audit cases. '
                    'Used only when OneView API audit is enabled. '
                    'See: https://github.com/openstack/python-oneviewclient'),
    cfg.StrOpt('audit_output_file',
               help='Path to OneView audit log file. '
                    'Created only when Oneview API audit is enabled.'),
]


def register_opts(conf):
    conf.register_opts(opts, group='oneview')



# Copyright 2016 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock
import unittest

from ironic_oneviewd import facade
from ironic_oneviewd.node_manager import manage
from ironic_oneviewd.node_manager.manage import NodeManager
from ironic_oneviewd import utils
from oneview_client.models import ServerProfile


class FakeIronicNode(object):
    def __init__(self, id, uuid, chassis_uuid, provision_state, driver,
                 ports, driver_info={}, driver_internal_info={},
                 name='fake-node', maintenance='False', properties={},
                 extra={}):

        self.id = id
        self.uuid = uuid
        self.chassis_uuid = chassis_uuid
        self.provision_state = provision_state
        self.driver = driver
        self.ports = ports
        self.driver_info = driver_info
        self.driver_internal_info = driver_internal_info
        self.maintenance = maintenance
        self.properties = properties
        self.extra = extra
        self.name = name


class FakeIronicPort(object):
    def __init__(self, id, uuid, node_uuid, address, extra={},
                 local_link_connection='', portgroup_id='',
                 pxe_enabled='False'):

        self.id = id
        self.uuid = uuid
        self.node_uuid = node_uuid
        self.address = address
        self.extra = extra
        self.local_link_connection = local_link_connection
        self.portgroup_id = portgroup_id
        self.pxe_enabled = pxe_enabled


class FakeServerHardware(object):
    def __init__(self, name, uuid, uri, power_state, port_map,
                 server_profile_uri, server_hardware_type_uri,
                 enclosure_group_uri, state, enclosure_uri):

        self.name = name
        self.uuid = uuid
        self.uri = uri
        self.power_state = power_state
        self.port_map = port_map
        self.server_profile_uri = server_profile_uri
        self.server_hardware_type_uri = server_hardware_type_uri
        self.enclosure_group_uri = enclosure_group_uri
        self.state = state
        self.enclosure_uri = enclosure_uri


class FakeConfHelper(object):
    def __init__(self, max_workers):
        self.rpc_thread_pool_size = max_workers


class FakeConfClient(object):
    def __init__(self, max_workers):
        self.DEFAULT = FakeConfHelper(max_workers)


POOL_OF_FAKE_IRONIC_NODES = [
    FakeIronicNode(
        id=123,
        uuid='66666666-7777-8888-9999-000000000000',
        chassis_uuid='aaaaaaaa-1111-bbbb-2222-cccccccccccc',
        maintenance=False,
        provision_state='enroll',
        ports=[
            {'id': 987,
             'uuid': '11111111-2222-3333-4444-555555555555',
             'node_uuid': '66666666-7777-8888-9999-000000000000',
             'address': 'AA:BB:CC:DD:EE:FF',
             'extra': {}}
        ],
        driver='fake_oneview',
        driver_info={'use_oneview_ml2_driver': True, 'user': 'foo',
                     'password': 'bar',
                     'server_hardware_uri':
                     '/rest/server-hardware-types/111112222233333'},
        properties={'num_cpu': 4},
        name='fake-node-1',
        extra={}
    ),
    FakeIronicNode(
        id=123,
        uuid='66666666-7777-8888-9999-000000000000',
        chassis_uuid='aaaaaaaa-1111-bbbb-2222-cccccccccccc',
        maintenance=False,
        provision_state='enroll',
        ports=[
            {'id': 987,
             'uuid': '11111111-2222-3333-4444-555555555555',
             'node_uuid': '66666666-7777-8888-9999-000000000000',
             'address': 'AA:BB:CC:DD:EE:FF',
             'extra': {}}
        ],
        driver='fake_oneview',
        driver_info={
            'dynamic_allocation': True,
            'user': 'foo', 'password': 'bar'
        },
        properties={'num_cpu': 4},
        name='fake-node-2',
        extra={}
    ),
    FakeIronicNode(
        id=123,
        uuid='66666666-7777-8888-9999-000000000000',
        chassis_uuid='aaaaaaaa-1111-bbbb-2222-cccccccccccc',
        maintenance=False,
        provision_state='enroll',
        ports=[
            {'id': 987,
             'uuid': '11111111-2222-3333-4444-555555555555',
             'node_uuid': '66666666-7777-8888-9999-000000000000',
             'address': 'AA:BB:CC:DD:EE:FF',
             'extra': {}}
        ],
        driver='fake_oneview',
        driver_info={'use_oneview_ml2_driver': False, 'user': 'foo',
                     'password': 'bar',
                     'server_hardware_uri':
                     '/rest/server-hardware-types/111112222233333'},
        properties={'num_cpu': 4},
        name='fake-node-3',
        extra={}
    ),
]

POOL_OF_FAKE_IRONIC_PORTS = [
    FakeIronicPort(
        id=987,
        uuid='11111111-2222-3333-4444-555555555555',
        local_link_connection=None,
        node_uuid='66666666-7777-8888-9999-000000000000',
        address='AA:BB:CC:DD:EE:FF',
        extra={}
    )
]

POOL_OF_FAKE_SERVER_HARDWARE = [
    FakeServerHardware(
        name='AAAAA',
        uuid='11111111-7777-8888-9999-000000000000',
        uri='/rest/server-hardware/11111',
        port_map={'deviceSlots': ''},
        power_state='Off',
        server_profile_uri='',
        server_hardware_type_uri='/rest/server-hardware-types/111112222233333',
        enclosure_group_uri='/rest/enclosure-groups/1111112222233333',
        state='Unknown',
        enclosure_uri='/rest/enclosures/1111112222233333',
    )
]


class TestIronicOneviewd(unittest.TestCase):
    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(facade.Facade, 'get_ironic_node_list')
    def test_take_node_actions(self, mock_get_ironic_node_list, mock_facade):

        mocked_facade = facade.Facade()
        mock_get_ironic_node_list.return_value = POOL_OF_FAKE_IRONIC_NODES
        mocked_facade.get_ironic_node_list = mock_get_ironic_node_list
        mock_facade.return_value = mocked_facade
        node_manager = NodeManager()
        node_manager.pull_ironic_nodes()
        mock_get_ironic_node_list.assert_called_with()

    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(manage.NodeManager, 'take_enroll_state_actions')
    def test_manage_node_provision_state_with_node_in_enroll(
        self, mock_take_enroll_state_actions, mock_facade
    ):
        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        fake_node.provision_state = 'enroll'
        node_manager = NodeManager()
        node_manager.manage_node_provision_state(fake_node)
        mock_take_enroll_state_actions.assert_called_with(fake_node)

    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(manage.NodeManager, 'take_manageable_state_actions')
    def test_manage_node_provision_state_with_node_in_manageable(
        self, mock_take_manageable_state_actions, mock_facade
    ):
        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        fake_node.provision_state = 'manageable'
        node_manager = NodeManager()
        node_manager.manage_node_provision_state(fake_node)
        mock_take_manageable_state_actions.assert_called_with(fake_node)

    @mock.patch('ironic_oneviewd.conf.CONF.openstack.inspection_enabled')
    @mock.patch('ironic_oneviewd.facade.Facade', new_callable=mock.MagicMock)
    def test_manage_provision_state_when_dynamic_allocation_inspection_enabled(
        self, mock_facade, mock_inspection_enabled
    ):
        mock_facade = facade.Facade()
        mock_facade.set_node_provision_state = mock.MagicMock()
        mock_inspection_enabled.return_value = True
        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[1])
        fake_node.provision_state = 'manageable'
        node_manager = NodeManager()
        node_manager.manage_node_provision_state(fake_node)
        mock_facade.set_node_provision_state.assert_called_with(fake_node,
                                                                'inspect')

    @mock.patch('ironic_oneviewd.facade.Facade', new_callable=mock.MagicMock)
    def test_manage_node_provision_state_with_node_in_inspect_failed(
        self, mock_facade
    ):
        mock_facade = facade.Facade()
        mock_facade.set_node_provision_state = mock.MagicMock()
        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[1])
        fake_node.provision_state = 'inspect failed'
        fake_node.last_error = ("OneView exception occurred. Error: Node %s is"
                                " already in use by OneView.") % fake_node.id
        node_manager = NodeManager()
        node_manager.manage_node_provision_state(fake_node)
        mock_facade.set_node_provision_state.assert_called_with(fake_node,
                                                                'manage')

    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(manage.NodeManager, 'is_bootable')
    def test_set_local_link_connection_use_ml2_oneview(
        self, mock_is_bootable, mock_facade
    ):
        mock_is_bootable.return_value = True
        switch_info = (
            '{"server_hardware_id": "111112222233333", "bootable": "True"}')

        local_link_connection = {
            "switch_id": "01:23:45:67:89:ab",
            "port_id": "",
            "switch_info": switch_info
            }

        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        node_manager = NodeManager()
        call_driver = node_manager.set_local_link_connection(
            fake_node, "AA:BB:CC:DD:EE:FF"
        )
        self.assertDictEqual(local_link_connection, call_driver)

    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    def test_set_local_link_connection_not_use_ml2_oneview(self, mock_facade):
        local_link_connection = {}
        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[2])
        node_manager = NodeManager()

        call_driver = node_manager.set_local_link_connection(
            fake_node, "AA:BB:CC:DD:EE:FF"
        )
        self.assertDictEqual(local_link_connection, call_driver)

    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(facade.Facade, 'get_port')
    @mock.patch.object(facade.Facade, 'get_server_hardware_state')
    @mock.patch.object(facade.Facade, 'get_server_hardware')
    @mock.patch.object(facade.Facade, 'get_port_list_by_mac')
    @mock.patch.object(facade.Facade, 'get_server_profile_assigned_to_sh')
    @mock.patch.object(facade.Facade, 'set_node_provision_state')
    @mock.patch.object(facade.Facade, 'create_node_port')
    @mock.patch.object(facade.Facade, 'generate_and_assign_sp_from_spt')
    def test_all_enroll_actions(
        self, mock_apply_server_profile, mock_create_node_port,
        mock_set_node_provision_state, mock_get_server_profile_assigned_to_sh,
        mock_get_port_list_by_mac, mock_get_server_hardware,
        mock_get_server_hardware_state, mock_get_port, mock_facade
    ):

        mocked_facade = facade.Facade()
        node_manager = NodeManager()

        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        fake_node.provision_state = 'enroll'

        info = {'server_hardware_uri': '/rest/server-hardware/123'}
        fake_node.driver_info = info
        mock.patch.dict(fake_node.driver_info, info, clear=True)

        properties = {'capabilities':
                      "server_profile_template_uri:"
                      "/rest/server-profile-templates/123"}
        fake_node.properties = properties
        mock.patch.dict(fake_node.properties, properties, clear=True)

        server_profile = ServerProfile()
        server_profile.connections = [
            {'mac': '01:23:45:67:89:ab', 'boot': {'priority': "primary"}}
        ]

        assigned_server_profile = server_profile
        mock_get_server_profile_assigned_to_sh.return_value = (
            assigned_server_profile
        )
        mocked_facade.get_server_profile_assigned_to_sh = (
            mock_get_server_profile_assigned_to_sh
        )

        server_hardware = copy.deepcopy(POOL_OF_FAKE_SERVER_HARDWARE[0])
        server_hardware_state = 'ProfileApplied'
        mock_get_server_hardware_state.return_value = server_hardware_state
        mock_get_server_hardware.return_value = server_hardware
        mocked_facade.get_server_hardware_state = (
            mock_get_server_hardware_state
        )
        mocked_facade.get_server_hardware = mock_get_server_hardware

        port_list_by_mac = []
        mock_get_port_list_by_mac.return_value = port_list_by_mac
        mocked_facade.get_port_list_by_mac = mock_get_port_list_by_mac

        fake_port = copy.deepcopy(POOL_OF_FAKE_IRONIC_PORTS[0])
        mock_get_port.return_value = fake_port
        mocked_facade.get_port = mock_get_port

        uri_server_profile_applied = '/rest/server-profiles/123'
        mock_apply_server_profile.return_value = uri_server_profile_applied
        mocked_facade.generate_and_assign_sp_from_spt = (
            mock_apply_server_profile
        )

        port_created = True
        mock_create_node_port.return_value = port_created
        mocked_facade.create_node_port = mock_create_node_port

        node_provision_state_changed = {'node': fake_node.uuid,
                                        'ex_msg': 'manageable'}
        mock_set_node_provision_state.return_value = (
            node_provision_state_changed
        )

        server_profile_template_uuid = utils.uuid_from_uri(
            uri_server_profile_applied
        )
        mocked_facade.set_node_provision_state = mock_set_node_provision_state

        node_manager.manage_node_provision_state(fake_node)

        mock_apply_server_profile.assert_called_with(
            'Ironic [%(uuid)s]' % {'uuid': fake_node.uuid},
            server_hardware.uuid,
            server_profile_template_uuid

        )

        mock_create_node_port.assert_called_with(
            fake_node.uuid,
            '01:23:45:67:89:ab'
        )

        mock_set_node_provision_state.assert_called_with(
            fake_node,
            'manage'
        )

    @mock.patch('ironic_oneviewd.node_manager.manage.LOG.warning')
    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(facade.Facade, 'get_server_hardware_state')
    @mock.patch.object(facade.Facade, 'get_server_profile_assigned_to_sh')
    @mock.patch.object(facade.Facade, 'get_server_hardware')
    @mock.patch.object(facade.Facade, 'create_node_port')
    @mock.patch.object(facade.Facade, 'generate_and_assign_sp_from_spt')
    def test_log_warning_when_hardware_already_has_server_profile_applied(
        self, mock_apply_server_profile, mock_create_node_port,
        mock_get_server_hardware, mock_get_server_profile_assigned_to_sh,
        mock_get_server_hardware_state, mock_facade, mock_log
    ):
        mocked_facade = facade.Facade()
        node_manager = NodeManager()

        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        fake_node.provision_state = 'enroll'

        info = {'server_hardware_uri': '/rest/server-hardware/123'}
        fake_node.driver_info = info
        mock.patch.dict(fake_node.driver_info, info, clear=True)

        properties = {'capabilities':
                      "server_profile_template_uri:"
                      "/rest/server-profile-templates/123"}
        fake_node.properties = properties
        mock.patch.dict(fake_node.properties, properties, clear=True)

        server_profile = ServerProfile()
        server_profile.name = 'Ironic [%s]' % fake_node.uuid
        server_profile.uri = '/rest/server-profiles/123'
        server_profile.connections = [
            {'mac': fake_node.ports[0]['address'], 'boot': {
                'priority': "primary"}}
        ]

        mock_get_server_profile_assigned_to_sh.return_value = server_profile
        mocked_facade.get_server_profile_assigned_to_sh = (
            mock_get_server_profile_assigned_to_sh
        )
        server_hardware_state = 'ProfileApplied'
        mock_get_server_hardware_state.return_value = server_hardware_state
        mocked_facade.get_server_hardware_state = (
            mock_get_server_hardware_state
        )

        node_manager.manage_node_provision_state(fake_node)

        msg = (
            'Node %s has a Server Profile /rest/server-profiles/123 assigned'
            % (fake_node.uuid)
        )

        mock_log.assert_called_with(msg)

    @mock.patch('ironic_oneviewd.node_manager.manage.LOG.warning')
    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(facade.Facade, 'get_server_hardware_state')
    @mock.patch.object(facade.Facade, 'get_server_profile_assigned_to_sh')
    @mock.patch.object(facade.Facade, 'get_server_hardware')
    @mock.patch.object(facade.Facade, 'create_node_port')
    @mock.patch.object(facade.Facade, 'generate_and_assign_sp_from_spt')
    def test_log_warning_when_port_configuration_throw_exception(
        self, mock_apply_server_profile, mock_create_node_port,
        mock_get_server_hardware, mock_get_server_profile_assigned,
        mock_get_server_hardware_state, mock_facade, mock_log
    ):
        mocked_facade = facade.Facade()
        node_manager = NodeManager()

        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        fake_node.provision_state = 'enroll'

        info = {'server_hardware_uri': '/rest/server-hardware/123'}
        fake_node.driver_info = info
        mock.patch.dict(fake_node.driver_info, info, clear=True)

        properties = {'capabilities':
                      "server_profile_template_uri:"
                      "/rest/server-profile-templates/123"}
        fake_node.properties = properties
        mock.patch.dict(fake_node.properties, properties, clear=True)

        server_profile = ServerProfile()
        server_profile.name = 'Ironic [%s]' % fake_node.uuid
        server_profile.uri = '/rest/server-profiles/123'
        server_profile.connections = [
            {'mac': '01:23:45:67:89:ab', 'boot': None}
        ]

        assigned_server_profile = server_profile
        mock_get_server_profile_assigned.return_value = assigned_server_profile
        mocked_facade.get_server_profile_assigned_to_sh = (
            mock_get_server_profile_assigned
        )

        server_hardware_state = 'ProfileApplied'
        mock_get_server_hardware_state.return_value = server_hardware_state
        mocked_facade.get_server_hardware_state = (
            mock_get_server_hardware_state
        )

        node_manager.manage_node_provision_state(fake_node)

        msg = (
            "No bootable connection was found for Server Profile "
            "/rest/server-profiles/123"
        )

        mock_log.assert_called_with(msg)

    @mock.patch('ironic_oneviewd.node_manager.manage.LOG.warning')
    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(facade.Facade, 'get_server_hardware_state')
    @mock.patch.object(facade.Facade, 'get_port')
    @mock.patch.object(facade.Facade, 'get_port_list_by_mac')
    @mock.patch.object(facade.Facade, 'get_server_profile_assigned_to_sh')
    @mock.patch.object(facade.Facade, 'get_server_hardware')
    @mock.patch.object(facade.Facade, 'create_node_port')
    @mock.patch.object(facade.Facade, 'generate_and_assign_sp_from_spt')
    def test_log_warning_when_node_already_has_port_for_mac_address(
        self, mock_apply_server_profile, mock_create_node_port,
        mock_get_server_hardware, mock_get_server_profile_assigned_to_sh,
        mock_get_port_list_by_mac, mock_get_port,
        mock_get_server_hardware_state, mock_facade, mock_log
    ):
        mocked_facade = facade.Facade()
        node_manager = NodeManager()

        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[0])
        fake_node.provision_state = 'enroll'

        info = {'server_hardware_uri': '/rest/server-hardware/123'}
        fake_node.driver_info = info
        mock.patch.dict(fake_node.driver_info, info, clear=True)

        properties = {'capabilities':
                      "server_profile_template_uri:"
                      "/rest/server-profile-templates/123"}
        fake_node.properties = properties
        mock.patch.dict(fake_node.properties, properties, clear=True)

        server_profile = ServerProfile()
        server_profile.connections = [
            {'mac': fake_node.ports[0]['address'], 'boot': {
                'priority': "primary"}}
        ]

        assigned_server_profile = server_profile
        mock_get_server_profile_assigned_to_sh.return_value = \
            assigned_server_profile
        mocked_facade.get_server_profile_assigned_to_sh = \
            mock_get_server_profile_assigned_to_sh
        server_hardware_state = 'ProfileApplied'
        mock_get_server_hardware_state.return_value = server_hardware_state
        mocked_facade.get_server_hardware_state = \
            mock_get_server_hardware_state

        port_list_by_mac = copy.deepcopy(POOL_OF_FAKE_IRONIC_PORTS)
        mock_get_port_list_by_mac.return_value = port_list_by_mac
        mocked_facade.get_port_list_by_mac = mock_get_port_list_by_mac

        fake_port = copy.deepcopy(POOL_OF_FAKE_IRONIC_PORTS[0])
        mock_get_port.return_value = fake_port
        mocked_facade.get_port = mock_get_port

        node_manager.manage_node_provision_state(fake_node)

        msg = (
            "A port with MAC address AA:BB:CC:DD:EE:FF was already "
            "created for this node"
        )

        mock_log.assert_called_with(msg)

        mock_get_port.assert_called_with(
            fake_port.uuid
        )

    @mock.patch('ironic_oneviewd.facade.Facade', autospec=True)
    @mock.patch.object(facade.Facade, 'get_port')
    @mock.patch.object(facade.Facade, 'get_server_hardware_state')
    @mock.patch.object(facade.Facade, 'get_port_list_by_mac')
    @mock.patch.object(facade.Facade, 'get_server_profile_assigned_to_sh')
    @mock.patch.object(facade.Facade, 'set_node_provision_state')
    @mock.patch.object(facade.Facade, 'create_node_port')
    @mock.patch.object(facade.Facade, 'get_server_hardware_mac')
    @mock.patch.object(facade.Facade, 'generate_and_assign_sp_from_spt')
    def test_all_enroll_actions_when_dynamic_allocation_flag_is_true(
        self, mock_apply_server_profile, mock_get_server_hardware_mac,
        mock_create_node_port, mock_set_node_provision_state,
        mock_get_server_profile_assigned_to_sh, mock_get_port_list_by_mac,
        mock_get_server_hardware_state, mock_get_port, mock_facade
    ):

        mocked_facade = facade.Facade()
        node_manager = NodeManager()

        fake_node = copy.deepcopy(POOL_OF_FAKE_IRONIC_NODES[1])
        fake_node.provision_state = 'enroll'

        info = {
            'dynamic_allocation': True,
            'server_hardware_uri': '/rest/server-hardware/123'
        }
        fake_node.driver_info = info
        mock.patch.dict(fake_node.driver_info, info, clear=True)

        properties = {'capabilities': ''}
        fake_node.properties = properties
        mock.patch.dict(fake_node.properties, properties, clear=True)

        server_profile = ServerProfile()
        server_profile.connections = [
            {'mac': '01:23:45:67:89:ab', 'boot': {'priority': "primary"}}
        ]

        mock_get_server_hardware_mac.return_value = '01:23:45:67:89:ab'
        mocked_facade.get_server_hardware_mac = mock_get_server_hardware_mac

        assigned_server_profile = server_profile
        mock_get_server_profile_assigned_to_sh.return_value = \
            assigned_server_profile
        mocked_facade.get_server_profile_assigned_to_sh = \
            mock_get_server_profile_assigned_to_sh

        server_hardware_state = 'ProfileApplied'
        mock_get_server_hardware_state.return_value = server_hardware_state
        mocked_facade.get_server_hardware_state = \
            mock_get_server_hardware_state

        port_list_by_mac = []
        mock_get_port_list_by_mac.return_value = port_list_by_mac
        mocked_facade.get_port_list_by_mac = mock_get_port_list_by_mac

        fake_port = copy.deepcopy(POOL_OF_FAKE_IRONIC_PORTS[0])
        mock_get_port.return_value = fake_port
        mocked_facade.get_port = mock_get_port

        uri_server_profile_applied = '/rest/server-profiles/123'
        mock_apply_server_profile.return_value = uri_server_profile_applied
        mocked_facade.generate_and_assign_sp_from_spt = \
            mock_apply_server_profile

        port_created = True
        mock_create_node_port.return_value = port_created
        mocked_facade.create_node_port = mock_create_node_port

        node_provision_state_changed = {'node': fake_node.uuid,
                                        'ex_msg': 'manageable'}
        mock_set_node_provision_state.return_value = \
            node_provision_state_changed
        mocked_facade.set_node_provision_state = mock_set_node_provision_state

        node_manager.manage_node_provision_state(fake_node)

        mock_apply_server_profile.assert_not_called()

        mock_create_node_port.assert_called_with(
            fake_node.uuid,
            '01:23:45:67:89:ab'
        )

        mock_set_node_provision_state.assert_called_with(
            fake_node,
            'manage'
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Daemon to the Ironic OneView drivers.
"""

from __future__ import print_function

import argparse
import six
import sys

from oslo_log import log as logging
from oslo_utils import encodeutils

from ironic_oneviewd.conf import CONF
from ironic_oneviewd.node_manager import commands as oneviewd_commands
from ironic_oneviewd.openstack.common._i18n import _
from ironic_oneviewd.openstack.common import cliutils

VERSION = '0.6.0'

COMMAND_MODULES = [
    oneviewd_commands,
]


class IronicOneViewD(object):

    def get_base_parser(self):
        parser = argparse.ArgumentParser(
            prog='ironic-oneviewd',
            description=__doc__.strip(),
            epilog='See "ironic-oneviewd --help COMMAND" '
                   'for help on a specific command.',
            add_help=False,
            formatter_class=HelpFormatter,
        )

        # Global arguments
        parser.add_argument('-h', '--help',
                            action='store_true',
                            help=argparse.SUPPRESS,
                            )

        parser.add_argument('--version',
                            action='version',
                            version=VERSION)

        parser.add_argument('-c', '--config-file',
                            default='/etc/ironic-oneviewd/'
                                    'ironic-oneviewd.conf',
                            help='Path to the Ironic OneView daemon '
                                 'configuration file.')

        parser.add_argument('-l', '--log-file',
                            help='Path to the Ironic OneView daemon '
                                 'logging file.')

        return parser

    def get_subcommand_parser(self, version):
        parser = self.get_base_parser()
        self.subcommands = {}
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        enhance_parser(parser, subparsers, self.subcommands)
        define_commands_from_module(subparsers, self, self.subcommands)
        return parser

    @cliutils.arg('command', metavar='<subcommand>', nargs='?',
                  help='Display help for <subcommand>')
    def do_help(self, args):
        """Display help about this program or one of its subcommands."""
        if getattr(args, 'command', None):
            if args.command in self.subcommands:
                self.subcommands[args.command].print_help()
            else:
                raise Exception(_("'%s' is not a valid subcommand") %
                                args.command)
        else:
            self.parser.print_help()

    def main(self, argv):
        log_domain = "DEFAULT"
        parser = self.get_base_parser()
        (options, args) = parser.parse_known_args(argv)
        subcommand_parser = self.get_subcommand_parser(1)
        self.parser = subcommand_parser

        logging.register_options(CONF)
        CONF(default_config_files=[options.config_file])
        logging.setup(CONF, log_domain)

        if options.help:
            self.do_help(options)

        if 'help' not in argv:
            oneviewd_commands.do_manage_ironic_nodes()

        args = subcommand_parser.parse_args(argv)

        # Short-circuit and deal with these node_manager right away.
        if args.func == self.do_help:
            self.do_help(args)
            return 0


def define_command(subparsers, command, callback, cmd_mapper):
    '''Define a command in the subparsers collection.

    :param subparsers: subparsers collection where the command will go
    :param command: command name
    :param callback: function that will be used to process the command
    '''
    desc = callback.__doc__ or ''
    help = desc.strip().split('\n')[0]
    arguments = getattr(callback, 'arguments', [])

    subparser = subparsers.add_parser(command, help=help,
                                      description=desc,
                                      add_help=False,
                                      formatter_class=HelpFormatter)
    subparser.add_argument('-h', '--help', action='help',
                           help=argparse.SUPPRESS)
    cmd_mapper[command] = subparser
    for (args, kwargs) in arguments:
        subparser.add_argument(*args, **kwargs)
    subparser.set_defaults(func=callback)


def define_commands_from_module(subparsers, command_module, cmd_mapper):
    """Add *do_* methods in a module and node_manager into a subparsers."""
    for method_name in (a for a in dir(command_module) if a.startswith('do_')):
        # Commands should be hypen-separated instead of underscores.
        command = method_name[3:].replace('_', '-')
        callback = getattr(command_module, method_name)
        define_command(subparsers, command, callback, cmd_mapper)


def enhance_parser(parser, subparsers, cmd_mapper):
    for command_module in COMMAND_MODULES:
        define_commands_from_module(subparsers, command_module, cmd_mapper)


class HelpFormatter(argparse.HelpFormatter):
    def start_section(self, heading):
        # Title-case the headings
        heading = '%s%s' % (heading[0].upper(), heading[1:])
        super(HelpFormatter, self).start_section(heading)


def main():
    try:
        IronicOneViewD().main(sys.argv[1:])
    except KeyboardInterrupt:
        print("\nironic-oneviewd stopped", file=sys.stderr)
        sys.exit(130)
    except Exception as e:
        print(encodeutils.safe_encode(six.text_type(e)), file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()
